import logging
import math
import numpy as np
from astropy.io import fits
from astropy.table import Table
from astropy import units as un
from . import archiveutils as au

"""
    Utility module for accessing FITS data
"""

logger = logging.getLogger('ingestion')

polarizations = {1 : 'I', 2 : 'Q', 3 : 'U', 4 : 'V'}


def get_fits_data(filename):
    """
    Get FITS header data
    :param filename:  name of the FITS file.
    :return: target name, ra, dec, image units, min frequency, max frequency
    """

    hdu = fits.open(filename)
    header = hdu['PRIMARY'].header
    image_data = hdu[0].data
    image_table = Table(image_data)
    image_array = image_table[0][0]

    source_name = get_source_name(header)
    ra = get_ra(header)
    dec = get_dec(header)
    image_field_of_view = get_image_field_of_view(header)

    ra_element_count = get_ra_element_count(header)
    dec_element_count = get_dec_element_count(header)
    spatial_resolution = get_spatial_resolution(header)
    beam_axis_ratio = get_beam_axis_ratio(header)

    min_frequency = get_min_frequency(header)
    max_frequency = get_max_frequency(header)

    rest_frequency = get_rest_frequency(header)
    polarization_id = get_polarization_id(header)

    telescope = get_telescope(header)
    image_units = get_image_units(header)
    min_intensity = get_min_intensity(image_array)
    max_intensity = get_max_intensity(image_array)
    rms_noise = get_rms_noise(image_array)

    spatial_region = get_spatial_region(header)

    ra_pixel_size = get_ra_pixel_size(header)
    dec_pixel_size = get_dec_pixel_size(header)

    return source_name, telescope, spatial_resolution, image_field_of_view, min_intensity, max_intensity, ra, dec, min_frequency, max_frequency,\
        rest_frequency, ra_element_count, dec_element_count, image_units, beam_axis_ratio, polarization_id, rms_noise, spatial_region,\
        ra_pixel_size, dec_pixel_size


def get_source_name(header):
    return header['OBJECT']


def get_ra(header):
    if header['CTYPE1'] == 'RA---SIN':
        return header['CRVAL1']
    else:
        return 0


def get_dec(header):
    if header['CTYPE2'] == 'DEC--SIN':
        return header['CRVAL2']
    else:
        return 0


def get_image_field_of_view(header):
    return math.sqrt((header['NAXIS1'] * header['CDELT1']) ** 2 + (header['NAXIS2'] * header['CDELT2']) ** 2)


def get_ra_element_count(header):
    if header['CTYPE1'] == 'RA---SIN':
        return header['NAXIS1']
    else:
        return 0


def get_dec_element_count(header):
    if header['CTYPE2'] == 'DEC--SIN':
        return header['NAXIS2']
    else:
        return 0


def get_spatial_resolution(header):
    """
    :param header:
    :return: spatial resolution in arcsec
    """
    spatial_resolution_degrees = (math.sqrt(header['BMAJ'] * header['BMIN'])) * un.degree
    return spatial_resolution_degrees.to(un.arcsec).value


def get_min_frequency(header):
    if header['CTYPE3'] == 'FREQ':
        return header['CRVAL3'] - header['CDELT3'] / 2
    else:
        return 0


def get_max_frequency(header):
    if header['CTYPE3'] == 'FREQ':
        return header['CRVAL3'] + header['CDELT3'] / 2
    else:
        return 0


def get_rest_frequency(header):
    return header['RESTFRQ']


def get_beam_axis_ratio(header):
    return header['BMAJ'] / header['BMIN']


def get_polarization_id(header):
    if header['CTYPE4'] == 'STOKES':
        fits_stokes = round(header['CRVAL4'])
        polarization_name = polarizations[fits_stokes]
        return au.getpolarizationcode(polarization_name)
    else:
        return 0


def get_telescope(header):
    return header['TELESCOP']


def get_image_units(header):
    return '{} {}'.format(header['BTYPE'], header['BUNIT'])


def get_min_intensity(image_array):
    return np.nanmin(image_array).astype(np.float64)


def get_max_intensity(image_array):
    return np.nanmax(image_array).astype(np.float64)


# TODO
def get_rms_noise(image_array):
    return 0


def get_spatial_region(header):
    # BOX IRCS {} {} {} {}".format(ra, dec, ra_extent, dec_extent)
    if header['CTYPE1'] == 'RA---SIN':
        ra_extent = abs(header['NAXIS1'] * header['CDELT1'])
    else:
        ra_extent = 0

    if header['CTYPE2'] == 'DEC--SIN':
        dec_extent = abs(header['NAXIS2'] * header['CDELT2'])
    else:
        dec_extent = 0

    return 'BOX IRCS {} {} {} {}'.format(get_ra(header), get_dec(header), ra_extent, dec_extent)


def get_ra_pixel_size(header):
    if header['CTYPE1'] == 'RA---SIN':
        return abs(header['CDELT1'])
    else:
        return 0


def get_dec_pixel_size(header):
    if header['CTYPE2'] == 'DEC--SIN':
        return abs(header['CDELT2'])
    else:
        return 0









