# -*- coding: utf-8 -*-

''' Project finders pull project metadata
    from the "new" and legacy archive DBs.
'''

import logging
import re
import warnings
from pprint import pprint
from sqlalchemy import exc as sa_exc

from schema import create_session, Project
from schema.legacy_model import LegacyProject
from support.logging import get_console_logger

# pylint: disable=logging-format-interpolation

class NewArchiveProjectFinder:
    ''' Pulls specified project from archive DB '''

    def __init__(self, profile: str):
        self.profile = profile
        self._LOG = get_console_logger(self.__class__.__name__)

    def find_project(self, project_code: str):
        ''' get specified project'''
        session = create_session('SDM', profile=self.profile)
        try:
            with warnings.catch_warnings():
                # suppress SQLAlchemy warnings
                warnings.simplefilter("ignore", category=sa_exc.SAWarning)

                # In the archive, the project code is a PK;
                # we should only  ever get one
                project = session.query(Project) \
                    .filter(Project.project_code == project_code) \
                    .first()

            return project

        except Exception as ex:
            self._LOG.error(f'{ex}')
            raise NewArchiveProjectException(ex)

        finally:
            session.close()


LEGACY_QUERY = """select PROJECT_CODE,
OBSERVER,
STARTTIME,
STOPTIME,
PROPRIETARY_DURATION,
STOPTIME + PROPRIETARY_DURATION,
proprietary,
UNLOCK_EXPIRE,
PROJECT_LOCK
from PROJECT
where PROJECT_CODE= :project_code 
order by PROPRIETARY desc"""


class LegacyProjectFinder:
    ''' Pulls specified project from legacy DB '''

    def __init__(self, profile: str):
        self.profile = profile
        self._LOG = get_console_logger(self.__class__.__name__)

    def find_projects(self, project_code: str) -> list:
        ''' project codes are not unique in the legacy archive '''

        legacy_project_code = make_legacy_project_code(project_code)
        session = create_session('LEGACY',
                                 profile=self.profile)
        try:

            with warnings.catch_warnings():
                # suppress SQLAlchemy warnings
                warnings.simplefilter("ignore", category=sa_exc.SAWarning)

                try:
                    projects = session.query(LegacyProject) \
                        .filter(LegacyProject.project_code
                                .in_((project_code, legacy_project_code))) \
                        .all()
                finally:
                    session.close()
            return projects

        except Exception as ex:
            self._LOG.error(f'{ex}')
            raise LegacyProjectException(ex)

        finally:
            session.close()

    def find_proprietary_metadata(self, project_code: str):
        ''' Get proprietary, lock, prop duration from legacy archive '''
        session = create_session('LEGACY',
                                 profile=self.profile)
        try:

            with warnings.catch_warnings():
                # suppress SQLAlchemy warnings
                warnings.simplefilter("ignore", category=sa_exc.SAWarning)

                cursor = session.connection().engine.raw_connection().cursor()
                cursor.execute(LEGACY_QUERY, project_code=project_code)

            results = list()
            for row in cursor.fetchall():
                (proj_code, operator, starttime, stoptime, prop_dur, stop_dur,
                 proprietary, unlock_expire, proj_lock) = row
                proj_metadata = {'project_code': proj_code,
                                 'operator': operator,
                                 'starttime': starttime,
                                 'stoptime': stoptime,
                                 'prop_duration': prop_dur,
                                 'stop_dur': stop_dur,
                                 'proprietary': proprietary,
                                 'unlock_expire': unlock_expire,
                                 'project_lock': proj_lock,
                                 }
                self._LOG.info(pprint(proj_metadata))
                results.append(proj_metadata)

            return results

        except Exception as ex:
            self._LOG.error(f'{ex}')
            raise LegacyProjectException(ex)

        finally:
            session.close()


def make_legacy_project_code(project_code: str):
    ''' Some project codes in the legacy DB have an extra 0 in them.
        Create such a project code from a given project code
        for use in searches.

     '''
    legacy_project_code = project_code.replace('0', '00', 1)
    if legacy_project_code != project_code:
        return legacy_project_code

    # project code didn't have a zero in it
    match = re.match(r"(([A-Za-z]+)([-0-9]+))", project_code)
    if match is None:
        return project_code
    parts = match.groups()
    if len(parts) == 3:
        return parts[1] + '0' + parts[2]

    logging.error(f"don't know what to do with {project_code}")
    return project_code


class NewArchiveProjectException(Exception):
    ''' raise this when something goes wrong in querying or updating
        the archive DB
    '''


class LegacyProjectException(Exception):
    ''' raise this when something goes wrong in querying or updating
        the legacy DB
    '''
