import zope.sqlalchemy
from pyramid.config import Configurator
from pyramid.renderers import JSONP
from pyramid.request import Request
from pyramid.view import view_config, view_defaults
from pyramid_beaker import session_factory_from_settings
from workspaces.services import WorkflowInfo, WorkflowService, get_session_factory, get_engine

"""
To consider:
 ☐ Do we want to use Marshmallow to validate inputs/generate API documentation?
 ☐ Do we want to have a separate package with interfaces/model? Or perhaps only JSON schemas?
"""


# ---------------------------------------------------------
#
#    S E R V I C E S
#
# ---------------------------------------------------------


def lookup_workflow(request):
    return request.info.lookup_workflow_definition(request.matchdict['name'])


def lookup_file(request):
    return lookup_workflow(request)['files'][request.matchdict['filename']]


@view_defaults(route_name='workflows', renderer='json')
class WorkflowRestService:
    """
    Top-level service for workflow requests.

    TODO: rename this to requests and add new service about workflow definitions with this name
    """
    def __init__(self, request: Request):
        self.request = request

    @view_config(request_method='GET')
    def list_workflows(self):
        """
        List the workflow requests that we know about.

        Audience: front-end
        :return:
        """
        return self.request.info.all_workflows()

    @view_config(request_method='POST', route_name='create_workflow')
    def create_workflow(self, name: str, args: str):
        """
        Create a new workflow request from the name/arguments supplied.

        Audience: front-end and CLI
        :return:
        """
        # all we should have to do here is take the WorkflowRequest from the context and
        # hand it to WorkflowInfo to save it, but we're still conflating
        # workflows and workflow requests right now
        req = self.request.context
        request = self.request.info.create_workflow_request(workflow_name=name, argument=args)
        return request

    @view_config(request_method='GET', route_name='workflow')
    def get_workflow(self):
        """
        Look up a workflow request

        Audience: front-end
        :return:
        """
        return self.request.context

    @view_config(request_method='POST', route_name='submit_workflow')
    def submit_workflow(self):
        """
        Submit this workflow request for processing.

        Audience: front-end and CLI
        :return:
        """
        print(f"Submitting workflow {self.request.context.workflow_name}")
        return self.request.workflows.execute(self.request.context)


@view_defaults(route_name='workflow_files', renderer='json')
class WorkflowFilesRestService:
    """
    Services for the user-submitted files attached to workflows.
    """
    def __init__(self, request):
        self.request = request

    @view_config(request_method='POST')
    def add_file(self):
        """
        Add a file to this workflow request.

        Audience: front-end and CLI
        """
        print('Adding a file')

    @view_config(request_method='GET')
    def get_files(self):
        """
        Get the files associated with this workflow request.

        Audience: front-end
        :return:
        """
        return self.request.context['files']

    @view_config(request_method='GET', route_name='workflow_file', accept='json', renderer='string')
    def get_file_text(self):
        """
        Get the text contents of this file

        Audience: ???
        :return:
        """
        return self.request.context['content']

    @view_config(request_method='GET', route_name='workflow_file', accept='application/json')
    def get_file_json(self):
        """
        Get a JSON representation of this file

        Audience: ???
        :return:
        """
        return self.request.context


# ---------------------------------------------------------
#
#    F R A M E W O R K    S E T U P
#
# ---------------------------------------------------------


def get_tm_session(session_factory, transaction_manager):
    """
    Enable Zope's transaction manager on our session
    :param session_factory:
    :param transaction_manager:
    :return:
    """
    dbsession = session_factory()
    zope.sqlalchemy.register(dbsession, transaction_manager=transaction_manager)
    return dbsession


# ---------------------------------------------------------
#
#    M A I N    E N T R Y P O I N T
#
# ---------------------------------------------------------


def main(global_config, **settings):
    with Configurator(settings=settings) as config:
        session_factory = session_factory_from_settings(settings)
        config.set_session_factory(session_factory)
        config.add_renderer('jsonp', JSONP(param_name='callback'))

        settings['tm.manager_hook'] = 'pyramid_tm.explicit_manager'

        # use pyramid_tm to hook the transaction lifecycle to the request
        config.include('pyramid_tm')

        # use pyramid_retry to retry a request when transient exceptions occur
        config.include('pyramid_retry')

        session_factory = get_session_factory(get_engine())
        config.registry['dbsession_factory'] = session_factory

        # make workflow_info available for use in Pyramid
        config.add_request_method(
            # r.tm is the transaction manager used by pyramid_tm
            lambda request: WorkflowInfo(get_tm_session(session_factory, request.tm)),
            'info',
            reify=True
        )
        # make workflow_service available for use in Pyramid
        config.add_request_method(lambda r: WorkflowService(r.info), 'workflows', reify=True)

        config.add_route('workflows', '/workflows')
        config.add_route('workflow', '/workflows/{name}', factory=lookup_workflow)
        config.add_route('create_workflow', '/workflows/create', factory=lookup_workflow)
        config.add_route('submit_workflow', '/workflows/{name}/submit', factory=lookup_workflow)
        config.add_route('workflow_files', '/workflows/{name}/files', factory=lookup_workflow)
        config.add_route('workflow_file', '/workflows/{name}/files/{filename}', factory=lookup_file)

        config.include('pyramid_beaker')
        config.scan('.')
        return config.make_wsgi_app()
